/*
 JungTaek Kim, The LET Project, KAIST

 Copyright(c) 2000 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*/

%{

open String_ast.Ast
open Location
open Syntaxerr

let syntax_error loc string = 
  raise (Error(Other(loc,string)))

let unclosed o_name o_num c_name c_num =
  raise(Error(Unclosed(Location.rhs_loc o_num, o_name,
                       Location.rhs_loc c_num, c_name)))

type sub = SUB | NOTSUB

let loc = Location.symbol_rloc

let sloc = Location.rhs_loc

let nloc = Location.none

let rloc s e = 
  let l = sloc s in
  let l' = sloc e in
    {Location.loc_start = l.Location.loc_start;
     Location.loc_end = l'.Location.loc_end;
     Location.loc_ghost = false}

let mkvarpat s i = VarPat((s,sloc i),sloc i)

let mkvarexp s i = VarExp(([],(s,sloc i),sloc i),sloc i)

let mktopexp l i = 
	Str(
	  SimpleDec(
	    ValDec([],
	      [(VarPat(("it",nloc),nloc),SeqExp(l,sloc i),sloc i)],
	      sloc i),
	    sloc i),
	  sloc i)
%}

/* Tokens */

%token ABSTYPE
%token AND
%token ANDALSO
%token AS
%token ASSIGN
%token CASE
%token DO
%token ELSE
%token END
%token EQTYPE
%token EXCEPTION
%token FN
%token FOR
%token FUN
%token FUNCTOR
%token HANDLE
%token IF
%token IN
%token INCLUDE
%token LET
%token LOCAL
%token OF
%token OP
%token OPEN
%token ORELSE
%token RAISE
%token REC
%token REF
%token SHARING
%token SIG
%token SIGNATURE
%token STRUCT
%token STRUCTURE
%token THEN
%token TYPE
%token VAL
%token WHERE
%token WHILE

%token NIL

%token <string> LID
%token <string> UID
%token <int> INT
%token <string> REAL
%token <string> STRING
%token <char> CHAR
%token <string> PATH

%token UNDERSCORE
%token LPAREN
%token RPAREN
%token LBRACE
%token RBRACE
%token COLON
%token SEMI
%token SEMISEMI
%token COMMA
%token STAR
%token ARROW
%token REVERSEARROW
%token DOUBLEARROW
%token BAR
%token LBRACKET
%token RBRACKET
%token LBRACKETBAR
%token BARRBRACKET
%token DOT
%token DOTDOTDOT
%token MINUS
%token MINUSMINUS
%token PLUS
%token PLUSPLUS
%token QUOTE
%token EQUAL
%token COLONCOLON
%token NOT
%token BANG
%token EOF

%token <string> PREFIX
%token <string> INFIX0
%token <string> INFIX1
%token <string> INFIX2
%token <string> INFIX3
%token <string> INFIX4
%token <string> INFIX5

/* Precedences and associativities. Lower precedences come first. */

%nonassoc prec_where
%left  AND
%right prec_fn prec_case prec_handle	/* match ... with ... */
%nonassoc SHARING
%nonassoc prec_one_seq
%nonassoc prec_seq
%nonassoc HANDLE
%right prec_raise                       /* raise */
%right BAR
%left SEMI                             /* e1; e2 (sequence) */
%nonassoc THEN
%nonassoc ELSE
%right  AS                               /* as in patterns */
%right ARROW				/* -> in type expressions */
%nonassoc  COLON
%left  COMMA                            /* , in expressions, patterns, types */
%right INFIX0 ASSIGN REVERSEARROW	/* assignments */
%left  ORELSE	                        /* or */
%left  ANDALSO				/* & */
%right NOT
%left  INFIX1 EQUAL                        
%right INFIX2
%right COLONCOLON
%left  INFIX3 PLUS MINUS                        
%left  INFIX4 STAR /* *, / */                        
%right INFIX5                           /* ** */ 
%nonassoc operator_arg
%nonassoc PATH
%left  DOT 				/* ** */ 
%right PREFIX REF PLUSPLUS MINUSMINUS BANG

/* Entry points */

%start batch_parse		/* for batch parsing (read from file) */
%type <String_ast.Ast.topdec> batch_parse
%start interactive_parse  /* for interactive parsing (read from keyboard) */
%type <String_ast.Ast.topdec> interactive_parse

%%

varid:	LID			{ ($1,loc()) } ;

opidpat:
    PREFIX			{ ($1,loc()) } 
  | INFIX0			{ ($1,loc()) }
  | INFIX1			{ ($1,loc()) }
  | INFIX2			{ ($1,loc()) }
  | INFIX3			{ ($1,loc()) }
  | INFIX4			{ ($1,loc()) }
  | INFIX5			{ ($1,loc()) }
  | PLUSPLUS			{ ("++",loc()) }
  | MINUSMINUS          { ("--",loc()) }
  | PLUS			{ ("+",loc()) }
  | MINUS			{ ("-",loc()) }
  | STAR			{ ("*",loc()) }
  | EQUAL			{ ("=",loc()) }
  | ANDALSO			{ ("andalso",loc()) }
  | ORELSE			{ ("orelse",loc()) }
  | NOT				{ ("not",loc()) }
   
opid:
    PLUSPLUS			{ ("++",loc()) }
  | MINUSMINUS          { ("--",loc()) }
  | PLUS			{ ("+",loc()) }
  | MINUS			{ ("-",loc()) }
  | STAR			{ ("*",loc()) }
  | EQUAL			{ ("=",loc()) }
  | ANDALSO			{ ("andalso",loc()) }
  | ORELSE			{ ("orelse",loc()) }
  | NOT				{ ("not",loc()) }
  | COLONCOLON			{ ("::",loc()) }
  | PREFIX			{ ($1,loc()) }
  | INFIX0			{ ($1,loc()) }
  | INFIX1			{ ($1,loc()) }
  | INFIX2			{ ($1,loc()) }
  | INFIX3			{ ($1,loc()) }
  | INFIX4			{ ($1,loc()) }
  | INFIX5			{ ($1,loc()) }
   
tyid:	
    LID				{ ($1,loc()) } 
  | REF 			{ ("ref",loc()) } ;

conid:	UID			{ ($1,loc()) } ;
strid:	UID			{ ($1,loc()) } ;
sigid:	UID			{ ($1,loc()) } ;
fctid:	UID			{ ($1,loc()) } ;

tyvar:	QUOTE LID		{ ($2,loc()) } ;

label:
    INT				{ (string_of_int( $1 ),loc()) }
  | LID				{ ($1,loc()) } ;

varlongid:
    varid 		{ ([],$1,loc()) }
  | PATH varlongid	{ match $2 with (x,y,_) -> (($1,sloc 1)::x,y,loc()) } ;

oplongid:
    PATH opid		{ ([($1,sloc 1)],$2,loc()) }
  | PATH oplongid	{ match $2 with (x,y,_) -> (($1,sloc 1)::x,y,loc()) }

tylongid:
    tyid 		{ ([],$1,loc()) }
  | PATH tylongid	{ match $2 with (x,y,_) -> (($1,sloc 1)::x,y,loc()) } ;

conlongid:
    conid		{ ([],$1,loc()) }
  | PATH conlongid	{ match $2 with (x,y,_) -> (($1,sloc 1)::x,y,loc()) } ;

strlongid:
    strid 		{ ([],$1,loc()) }
  | PATH strlongid	{ match $2 with (x,y,_) -> (($1,sloc 1)::x,y,loc()) } ;

sty:
    tyvar			{ VarTy($1,loc()) }
  | tylongid			{ ConstTy([],$1,loc()) }
  | sty tylongid		{ ConstTy([$1],$2,loc()) }
  | tyarg tylongid		{ ConstTy($1,$2,loc()) } 
  | LPAREN ty RPAREN 		{ $2 }
  | LPAREN error 		{ unclosed "(" 1 ")" 2;RecordTy([],loc()) }
  | LBRACE tyrow RBRACE		{ RecordTy($2,loc()) } ;
  | LBRACE error		{ unclosed "{" 1 "}" 2;RecordTy([],loc()) } ;

ty:
    sty				{ $1 }
  | tytup			{ TupleTy($1,loc()) } 
  | ty ARROW ty			{ FunTy($1,$3,loc()) } ;  

tytup:
    sty STAR sty		{ [$1;$3] }
  | tytup STAR sty		{ $1@[$3] } ;

tyarg:
    LPAREN tycommas RPAREN	{ $2 } ;

tycommas:
    ty COMMA ty			{ [$1;$3] }
  | tycommas COMMA ty 		{ $1@[$3] } ;

tyrow:
    label COLON ty		{ [($1,$3,loc())] }
  | tyrow COMMA label COLON ty	{ $1@[($3,$5,rloc 3 5)] } ;

spat:
    UNDERSCORE			{ WildPat(loc()) }
  | INT				{ IntPat($1,loc()) }
  | MINUS INT			{ IntPat(-$2,loc()) }
  | PLUS INT			{ IntPat($2,loc()) }
  | STRING			{ StringPat($1,loc()) }
  | CHAR			{ CharPat($1,loc()) }
  | NIL				{ ListPat([],loc()) }
  | LPAREN RPAREN		{ UnitPat(loc()) }
  | LPAREN error		{ unclosed "(" 1 ")" 2;WildPat(loc()) }
  | LPAREN opidpat RPAREN	{ VarPat($2,loc()) }
  | LPAREN patcommas RPAREN	
  			{ match $2 with [p] -> p | l -> TuplePat(l,loc()) }
  | LBRACE subpatrow RBRACE	{ match $2 with
                                    (p,SUB) -> SubRecordPat(p,loc())
				  | (p,NOTSUB) -> RecordPat(p,loc()) }
  | LBRACE error		{ unclosed "{" 1 "}" 2;WildPat(loc()) }
  | varid			{ VarPat($1,loc()) }
  | conlongid			{ ConPat($1,loc()) }
  | LBRACKET RBRACKET		{ ListPat([],loc()) }
  | LBRACKET patcommas RBRACKET	{ ListPat($2,loc()) }
  | LBRACKET error		{ unclosed "[" 1 "]" 2;WildPat(loc()) }
  | LBRACKETBAR patcommas BARRBRACKET
  				{ ArrayPat($2,loc()) } ;
  | LBRACKETBAR error		{ unclosed "[|" 1 "|]" 2;WildPat(loc()) } ;

apat:
    spat			{ $1 }
  | REF spat			{ RefPat($2,loc()) }
  | conlongid spat		{ AppPat($1,$2,loc()) }
  | LPAREN COLONCOLON RPAREN spat	
  	{ AppPat(([],("::",sloc 2),sloc 2),$4,loc()) }
  | apat COLONCOLON apat	
	{ AppPat(([],("::",sloc 2),sloc 2),TuplePat ([$1;$3],loc()),loc()) }
  | apat COLON ty		{ ConstraintPat($1,$3,loc()) } ;

aspat:
    apat			{ $1 }
  | apat AS apat		{ match $1 with
				    VarPat(v,_) -> AsPat(v,None,$3,loc())
				  | ConstraintPat(VarPat(v,_),t,_) ->
				      AsPat(v,Some(t),$3,loc()) 
				  | _ -> syntax_error (sloc 1) IllegalAsPat }
				  /* Errorcheck must check $1 is a variable */ 

orpat:
    aspat BAR aspat		{ [$1;$3] }
  | orpat BAR aspat		{ $1@[$3] } ;

pat:
    aspat			{ $1 }
  | orpat			{ OrPat($1,loc()) } ;

subpatrow:
    DOTDOTDOT			{ ([],SUB) } 
  | patrow			{ $1 } 
  | patrow COMMA DOTDOTDOT	{ match $1 with (x,y) -> (x,SUB) } ;

patrow:
    label EQUAL pat		{ ([($1,$3,loc())],NOTSUB) }
  | LID 		{ ([(($1,loc()),mkvarpat $1 1,loc())],NOTSUB) }
  | patrow COMMA label EQUAL pat{ match $1 with
				  (x,y) -> (x@[($3,$5,rloc 3 5)],y) } 
  | patrow COMMA LID { match $1 with
				  (x,y) -> (x@[(($3,loc()),mkvarpat $3 3,loc())],y) } 
  | patrow COMMA error	{ syntax_error (sloc 3) IllegalLabel }
  | patrow COMMA label error	
  			{ syntax_error (rloc 3 4) IllegalLabel }
  | patrow COMMA label EQUAL error
  			{ syntax_error (sloc 5) IllegalField } ;

patcommas:
    pat				{ [$1] }
  | patcommas COMMA pat		{ $1@[$3] } ;
  | patcommas COMMA error	{ syntax_error (sloc 3) IllegalPattern } ;

sexp:
    INT				{ IntExp($1,loc()) }
  | REAL			{ RealExp($1,loc()) }
  | STRING			{ StringExp($1,loc()) }
  | CHAR			{ CharExp($1,loc()) }
  | NIL				{ ListExp([],loc()) }
  | LPAREN RPAREN		{ UnitExp(loc()) }
  | LPAREN expseq RPAREN	{ SeqExp($2,loc()) }
  | LPAREN error		{ unclosed "(" 1 ")" 2;ListExp([],loc()) }
  | LPAREN expcommas RPAREN 	{ TupleExp($2,loc()) }
  | LBRACE exprow RBRACE	{ RecordExp($2,loc()) }
  | LBRACE error		{ unclosed "{" 1 "}" 2;ListExp([],loc()) }
  | LBRACKET RBRACKET		{ ListExp([],loc()) }
  | LBRACKET expseq RBRACKET	{ ListExp([SeqExp($2,sloc 2)],loc()) }
  | LBRACKET expcommas RBRACKET	{ ListExp($2,loc()) }
  | LBRACKET error		{ unclosed "[" 1 "]" 2;ListExp([],loc()) }
  | LBRACKETBAR expseq BARRBRACKET	{ ArrayExp([SeqExp($2,sloc 2)],loc()) }
  | LBRACKETBAR expcommas BARRBRACKET	{ ArrayExp($2,loc()) }
  | LBRACKETBAR error		{ unclosed "[|" 1 "|]" 2;ListExp([],loc()) }  
  | WHILE expseq DO expseq END	
  			{ WhileExp(SeqExp($2,sloc 2),SeqExp($4,sloc 4),loc()) }
  | WHILE expseq DO expseq error{ unclosed "do" 3 "end" 5;ListExp([],loc()) } 
  | FOR varid EQUAL exp SEMI exp SEMI exp DO expseq END	
			{ ForExp($2,$4,$6,$8,SeqExp($10,sloc 10),loc()) }
  | FOR varid EQUAL exp SEMI exp SEMI exp DO expseq error	
				{ unclosed "do" 9 "end" 11;ListExp([],loc()) } 
  | LET decseq IN expseq END	
  			{ LetExp(SeqDec($2,sloc 2),SeqExp($4,sloc 4),loc()) } 
  | LET decseq IN expseq error	{ unclosed "in" 3 "end" 5;ListExp([],loc()) } 
  | varlongid			{ VarExp($1,loc()) }
  | conlongid			{ ConExp($1,loc()) } 
  | oplongid			{ VarExp($1,loc()) }
  | LPAREN opid RPAREN	{ VarExp(([],$2,sloc 2),loc()) }
  | LPAREN opid error	{ unclosed "(" 1 ")" 3;ListExp([],loc()) } 
  | sexp PLUSPLUS		{ AppExp(mkvarexp "++" 2,$1,loc()) }
  | sexp MINUSMINUS		{ AppExp(mkvarexp "--" 2,$1,loc()) }
  | PREFIX sexp			{ AppExp(mkvarexp $1 1,$2,loc()) }
  | REF sexp			{ RefExp($2,loc()) }
  | BANG sexp			{ DeRefExp($2,loc()) }
  | sexp DOT label		{ RecordFieldExp($1,$3,loc()) }
  | sexp DOT LBRACKET expseq RBRACKET	
  				{ ArrayFieldExp($1,SeqExp($4,sloc 4),loc()) }
  | sexp DOT LBRACKET expseq error	{ unclosed "[" 3 "]" 5;ListExp([],loc()) } 
  | NOT sexp			{ AppExp(mkvarexp "not" 1,$2,loc()) }
  | PLUS sexp			{ $2 }
  | MINUS sexp	{ match $2 with
                    IntExp(i,_) -> IntExp(-i,loc())
                  | x -> AppExp(mkvarexp "unary_minus" 1,x,loc()) }
aexp:
    sexp			{ $1 }
  | aexp sexp			{ AppExp($1, $2,loc()) }
  | aexp LBRACE label REVERSEARROW expseq RBRACE	
			{ SubstRecordExp($1,$3,SeqExp($5,sloc 5),loc()) }
  | aexp LBRACE label REVERSEARROW expseq error
			{ unclosed "{" 2 "}" 6;ListExp([],loc()) } 
opexp:
    aexp			{ $1 } %prec operator_arg 
  | opexp INFIX1 opexp	
  	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp INFIX2 opexp	
 	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp INFIX3 opexp	
  	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp INFIX4 opexp	
  	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp INFIX5 opexp	
  	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp ANDALSO opexp	
  	{ AppExp(mkvarexp "andalso" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp ORELSE opexp	
  	{ AppExp(mkvarexp "orelse" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp COLONCOLON opexp
  	{ AppExp(mkvarexp "::" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp PLUS opexp	
  	{ AppExp(mkvarexp "+" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp MINUS opexp	
  	{ AppExp(mkvarexp "-" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp STAR opexp	
  	{ AppExp(mkvarexp "*" 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp EQUAL opexp	
  	{ AppExp(mkvarexp "=" 2,TupleExp([$1;$3],loc()),loc()) }
  | sexp DOT LBRACKET expseq RBRACKET REVERSEARROW opexp
			{ UpdateArrayExp($1,SeqExp($4,sloc 4),$7,loc()) } 
  | opexp INFIX0 opexp		
  	{ AppExp(mkvarexp $2 2,TupleExp([$1;$3],loc()),loc()) }
  | opexp ASSIGN opexp		
  	{ AssignExp($1,$3,loc()) }

exp:
    opexp			{ $1 }
  | exp HANDLE matchlist	{ HandleExp($1,$3,loc()) }	
  				%prec prec_handle
  | RAISE exp			{ RaiseExp($2,loc()) }	%prec prec_raise
  | FN fnmatchlist		{ FnExp($2,loc()) }	%prec prec_fn
  | CASE expseq OF matchlist	{ CaseExp(SeqExp($2,sloc 2),$4,loc()) }	
  				%prec prec_case
  | IF expseq THEN exp	
  			{ IfExp(SeqExp($2,sloc 2),$4,UnitExp(nloc),loc()) }
  | IF expseq THEN exp ELSE exp	
			{ IfExp(SeqExp($2,sloc 2),$4,$6,loc()) }
  | exp COLON ty		{ ConstraintExp($1,$3,loc()) } ;

expseq:
    exp 			{ [$1] }	%prec prec_one_seq
  | exp SEMI expseq		{ $1::$3 } ;
  | exp SEMI error		{ syntax_error (sloc 3) IllegalExpSeq } ;

expcommas:
    expseq COMMA expseq		{ [SeqExp($1,sloc 1);SeqExp($3,sloc 3)] }
  | expcommas COMMA expseq	{ $1@[SeqExp($3,sloc 3)] } ;

exprow:
    label EQUAL exp		{ [($1,$3,loc())] }
  | exprow COMMA label EQUAL exp	{ $1@[($3,$5,rloc 3 5)] } ;

fnmatchlist:
    spats DOUBLEARROW expseq	{ [($1,SeqExp($3,sloc 3),loc())] }
  | fnmatchlist BAR spats DOUBLEARROW expseq	
  				{ $1@[($3,SeqExp($5,sloc 5),rloc 3 5)] } ;
matchlist:
    pat DOUBLEARROW expseq	{ [($1,SeqExp($3,sloc 3),loc())] }
  | matchlist BAR pat DOUBLEARROW expseq	
  				{ $1@[($3,SeqExp($5,sloc 5),rloc 3 5)] } ;

dec:
    VAL valbinds		{ ValDec([],$2,loc()) }
  | VAL tyvseq valbinds		{ ValDec($2,$3,loc()) }
  | VAL REC valbinds		{ RecValDec([],$3,loc()) }
  | VAL REC tyvseq valbinds	{ RecValDec($3,$4,loc()) }
  | FUN funbinds		{ FunDec([],$2,loc()) }
  | FUN tyvseq funbinds		{ FunDec($2,$3,loc()) }
  | TYPE typbinds		{ TypeDec($2,loc()) }
  | ABSTYPE typbinds IN decseq END
  		{ AbstypeDec($2,SeqDec($4,sloc 4),loc()) }
  | ABSTYPE typbinds IN decseq error	
  		{ unclosed "abstype" 1 "end" 5;SeqDec([],loc()) }
  | EXCEPTION exnbinds		{ ExceptionDec($2,loc()) }
  | OPEN strlongids		{ OpenDec($2,loc()) }
  | LOCAL decseq IN decseq END	
  		{ LocalDec(SeqDec($2,sloc 2),SeqDec($4,sloc 4),loc()) }
  | LOCAL decseq IN decseq error	
  		{ unclosed "in" 3 "end" 5;SeqDec([],loc()) } ;

strlongids:
    strlongid			{ [$1] }
  | strlongids strlongid	{ $1@[$2] } ;

decseqelm:
    dec				{ $1 }
  | dec	SEMI			{ $1 } ;

decseq:
    decseqelm 			{ [$1] }
  | decseq decseqelm		{ $1@[$2] } ;

strdec:
    dec				{ SimpleDec($1,loc()) }
  | STRUCTURE strbinds		{ StrDec($2, loc()) } ;

strdecseqelm:
    strdec			{ $1 }
  | strdec	SEMI		{ $1 } ;

strdecseq:
    /* empty */ 		{ [] }
  | strdecseq strdecseqelm	{ $1@[$2] } ;

fctdec:
    FUNCTOR fctid LPAREN fctarg RPAREN sigexpop EQUAL strexp
    				{ ($2,$4,$6,$8,loc()) }
  | FUNCTOR fctid LPAREN fctarg error
    				{ unclosed "(" 3 ")" 6;(("error",loc()),[],None,StrStr(SeqStrDec([],loc()),loc()),loc()) } ;

sigexpop:
  				{ None }
  | COLON sigexp		{ Some($2) } ;

fctarg:
    strid COLON sigexp			{ [($1,$3,loc())] }
  | fctarg COMMA strid COLON sigexp	{ $1@[($3,$5,rloc 3 5)] } ;

sigdec:
    SIGNATURE sigbinds		{ ($2,loc()) } ;

strexp:
    strlongid			{ VarStr($1,loc()) }
  | STRUCT strdecseq END	{ StrStr(SeqStrDec($2,sloc 2),loc()) }
  | STRUCT strdecseq error	{ unclosed "struct" 1 "end" 3;StrStr(SeqStrDec([],loc()),loc()) }
  | strexp COLON sigexp			{ SigStr($1,$3,loc()) }
  | fctid LPAREN strexpseq RPAREN	{ FctAppStr($1,$3,loc()) }
  | fctid LPAREN strexpseq error	{ unclosed "(" 2 ")" 4;StrStr(SeqStrDec([],loc()),loc()) } ;

strexpseq:
    strexp			{ [$1] }
  | strexpseq COMMA strexp	{ $1@[$3] } ;

valbinds:
    pat EQUAL exp		{ [($1,$3,loc())] }
  | valbinds AND pat EQUAL exp	{ $1@[($3,$5,rloc 3 5)] } ;

bodies:
    varid spats EQUAL exp		{ [($1,$2,$4,loc())] }
  | LPAREN opid RPAREN spats EQUAL exp	{ [($2,$4,$6,loc())] }
  | bodies BAR varid spats EQUAL exp	{ $1@[($3,$4,$6,rloc 3 6)] } ;
  | bodies BAR LPAREN opid RPAREN spats EQUAL exp	
  					{ $1@[($4,$6,$8,rloc 4 8)] } ;

spats:
    spat			{ [$1] }
  | spats spat			{ $1@[$2] } ;

funbinds:
    bodies			{ [($1,loc())] }
  | funbinds AND bodies		{ $1@[($3,sloc 3)] } ;

typbind:
    tyvseqop tyid EQUAL ty		{ [TypeBind($1,$2,$4,loc())] }
  | tyvseqop tyid EQUAL conbinds	{ [DataBind($1,$2,$4,loc())] }

typbinds:
    typbind			{ $1 }
  | typbinds AND typbind	{ $1@$3 } ;

tyvseqop:
    /* empty */ 		{ [] }
  | tyvar			{ [$1] }
  | LPAREN tyvarcommas RPAREN	{ $2 }
  | LPAREN tyvarcommas error	{ unclosed "(" 1 ")" 3;[] } ;

tyvseq:
    tyvar			{ [$1] }
  | LPAREN tyvarcommas RPAREN	{ $2 } 
  | LPAREN tyvarcommas error	{ unclosed "(" 1 ")" 3;[] } ;

tyvarcommas:
    tyvar 			{ [$1] }
  | tyvarcommas COMMA tyvar	{ $1@[$3] } ;

conbind:
    conid 			{ ($1,None,loc()) }
  | conid OF ty			{ ($1,Some($3),loc()) } ;

conbinds:
    conbind			{ [$1] }
  | conbinds BAR conbind	{ $1@[$3] } ;

exnbind:
    conid 			{ ($1,None,loc()) }
  | conid OF ty			{ ($1,Some($3),loc()) } ;

exnbinds:
    exnbind			{ [$1] }
  | exnbinds AND exnbind	{ $1@[$3] } ;

sigbinds:
    sigid EQUAL sigexp			{ [($1,$3,loc())] }
  | sigbinds AND sigid EQUAL sigexp	{ $1@[($3,$5,rloc 3 5)] } ;

strbind:
    strid EQUAL strexp			{ ($1,None,$3,loc()) }
  | strid COLON sigexp EQUAL strexp	{ ($1,Some($3),$5,loc()) } ;

strbinds:
    strbind			{ [$1] }
  | strbinds AND strbind	{ $1@[$3] } ;

sigexp:
    sigid			{ VarSig($1,loc()) }
  | SIG END			{ SigSig(SeqSpec([],sloc 2),loc()) }
  | SIG specseq END		{ SigSig(SeqSpec($2,sloc 2),loc()) }
  | SIG specseq error	
  		{ unclosed "sig" 1 "end" 3;SigSig(SeqSpec([],loc()),loc()) }
  | sigexp where		{ ConstraintSig($1,$2,loc()) } ;

specseq:
    spec			{ [$1] } 	%prec prec_one_seq
  | spec SEMI specseq		{ $1::$3 }
  | spec specseq		{ $1::$2 } 	%prec prec_seq ;

spec:
    VAL valdescs		{ ValSpec($2,loc()) }
  | TYPE typdescs		{ TypeSpec($2,loc()) }
  | EQTYPE typdescs		{ TypeSpec($2,loc()) }
  | EXCEPTION exndescs		{ ExnSpec($2,loc()) }
  | INCLUDE sigexp		{ IncludeSpec($2,loc()) }
  | STRUCTURE strdescs		{ StrSpec($2,loc()) }
/*
  | specseq sharing	{ ConstraintSpec(SeqSpec($1,sloc 1),$2,loc()) } ;
*/

valdescs:
    varid COLON ty		{ [($1,$3,loc())] }
  | LPAREN opid RPAREN COLON ty	{ [($2,$5,loc())] }
  | valdescs AND varid COLON ty { $1@[($3,$5,rloc 3 5)] } ;
  | valdescs AND LPAREN opid RPAREN COLON ty 	
  				{ $1@[($4,$7,rloc 4 7)] } ;

typdesc:
    tyvseqop tyid 		{ [TypeDesc($1,$2,loc())] }
  | tyvseqop tyid EQUAL ty		{ [TypeBindDesc($1,$2,$4,loc())] }
  | tyvseqop tyid EQUAL condescs	{ [DataDesc($1,$2,$4,loc())] }

typdescs:
    typdesc			{ $1 }
  | typdescs AND typdesc	{ $1@$3 } ;

condescs:
    condesc			{ [$1] }
  | condescs BAR condesc	{ $1@[$3] } ;

condesc:
    conid 			{ ($1,None,loc()) }
  | conid OF ty			{ ($1,Some($3),loc()) } ;

exndescs:
    exndesc			{ [$1] }
  | exndescs AND exndesc	{ $1@[$3] } ;

exndesc:
    conid 			{ ($1,None,loc()) }
  | conid OF ty			{ ($1,Some($3),loc()) } ;

strdescs:
    strid COLON sigexp		{ [($1,$3,loc())] }
  | strdescs AND strid COLON sigexp 	{ $1@[($3,$5,rloc 3 5)] } ;
    
/*
sharing: SHARING TYPE longideqs	{ Sharing($3,loc()) }

longideqs:
    tylongid EQUAL tylongid	{ [LongIdEq($1,$3,loc())] }
  | longideqs AND tylongid EQUAL tylongid 	
  				{ $1@[LongIdEq($3,$5,rloc 3 5)] } ;
*/

where:
    WHERE TYPE longtypebinds	{ ($3,loc()) }	%prec prec_where
  | LPAREN WHERE TYPE longtypebinds RPAREN	{ ($4,loc()) } ;

longtypebinds:
    tyvseqop tylongid EQUAL ty	{ [($1,$2,$4,loc())] }
  | longtypebinds AND tyvseqop tylongid EQUAL ty 
  				{ $1@[($3,$4,$6,rloc 3 6)] } ;

topdec:
    strdec			{ Str($1,loc()) }
  | fctdec			{ Fct($1,loc()) }
  | sigdec			{ Sig($1,loc()) } ;

topdecselm:
    topdec			{ $1 }
  | topdec SEMI			{ $1 } ;

topdecs:
    topdecselm			{ [$1] }
  | topdecs topdecselm		{ $1@[$2] } ;

topdecseqtail:
				{ [] }
  | SEMISEMI			{ [] } ;
  | SEMISEMI expseq topdecseqtail	{ (mktopexp $2 1)::$3 }
  | SEMISEMI topdec topdecseqtail	{ $2::$3 } ;
  | SEMI topdec topdecseqtail	{ $2::$3 } ;
  | topdec topdecseqtail	{ $1::$2 } ;

topdecseq:
    topdecseqtail		{ $1 }
  | expseq topdecseqtail	{ (mktopexp $1 1)::$2 }

/* Entry points */

batch_parse:
    topdecseq EOF		{ SeqTopDec($1,loc()) } ;

interactive_parse:
    topdecs SEMISEMI		{ SeqTopDec($1,loc()) } 
  | expseq SEMISEMI		{ mktopexp $1 1 } 
  | EOF				{ raise End_of_file } ; 
    
%%

(*
 Modification history.

 2000/03/27 By Judaigi

 - Create typbind and typdesc rules.
 - Change typbinds and typdescs rules according to typbind and typdesc rules.
 - Change according to changes in ast.
*)
